library align_positioned;

import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';

////////////////////////////////////////////////////////////////////////////////////////////////////

enum Touch { inside, outside }

////////////////////////////////////////////////////////////////////////////////////////////////////

class AlignPositioned extends SingleChildRenderObjectWidget {
  //
  final Alignment alignment;
  final double dx, dy;
  final double childWidth, childHeight;
  final double containerWidth, containerHeight;
  final Touch touch;

  const AlignPositioned({
    Key key,
    this.alignment = Alignment.center,
    this.dx = 0.0,
    this.dy = 0.0,
    this.childWidth = 0.0,
    this.childHeight = 0.0,
    this.containerWidth = 0.0,
    this.containerHeight = 0.0,
    this.touch = Touch.inside,
    Widget child,
  })  : assert(alignment != null),
        super(key: key, child: child);

  @override
  _RenderAlignPositionedBox createRenderObject(BuildContext context) {
    return _RenderAlignPositionedBox(
      alignment: alignment,
      dx: dx ?? 0.0,
      dy: dy ?? 0.0,
      childWidth: childWidth ?? 0.0,
      childHeight: childHeight ?? 0.0,
      containerWidth: containerWidth ?? 0.0,
      containerHeight: containerHeight ?? 0.0,
      touch: touch,
      textDirection: Directionality.of(context),
    );
  }

  @override
  void updateRenderObject(BuildContext context, _RenderAlignPositionedBox renderObject) {
    renderObject
      ..alignment = alignment
      ..dx = dx ?? 0.0
      ..dy = dy ?? 0.0
      ..childWidth = childWidth ?? 0.0
      ..childHeight = childHeight ?? 0.0
      ..containerWidth = containerWidth ?? 0.0
      ..containerHeight = containerHeight ?? 0.0
      ..touch = touch;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////

class _RenderAlignPositionedBox extends RenderShiftedBox {
  //
  _RenderAlignPositionedBox({
    RenderBox child,
    @required double dx,
    @required double dy,
    @required double childWidth,
    @required double childHeight,
    @required double containerWidth,
    @required double containerHeight,
    @required Touch touch,
    Alignment alignment = Alignment.center,
    TextDirection textDirection,
  })  : assert(dx != null),
        assert(dy != null),
        assert(touch != null),
        _dx = dx,
        _dy = dy,
        _childWidth = childWidth,
        _childHeight = childHeight,
        _containerWidth = containerWidth,
        _containerHeight = containerHeight,
        _alignment = alignment,
        _touch = touch,
        super(child);

  // ---

  Alignment get alignment => _alignment;
  Alignment _alignment;

  set alignment(Alignment value) {
    assert(value != null);
    if (_alignment == value) return;
    _alignment = value;
    markNeedsLayout();
  }

  // ---

  Touch get touch => _touch;
  Touch _touch;

  set touch(Touch value) {
    assert(value != null);
    if (_touch == value) return;
    _touch = value;
    markNeedsLayout();
  }

  // ---

  double get childWidth => _childWidth;
  double _childWidth;

  set childWidth(double value) {
    assert(value != null);
    if (_childWidth == value) return;
    _childWidth = value;
    markNeedsLayout();
  }

  // ---

  double get childHeight => _childHeight;
  double _childHeight;

  set childHeight(double value) {
    assert(value != null);
    if (_childHeight == value) return;
    _childHeight = value;
    markNeedsLayout();
  }

  // ---

  double get containerWidth => _containerWidth;
  double _containerWidth;

  set containerWidth(double value) {
    assert(value != null);
    if (_containerWidth == value) return;
    _containerWidth = value;
    markNeedsLayout();
  }

  // ---

  double get containerHeight => _containerHeight;
  double _containerHeight;

  set containerHeight(double value) {
    assert(value != null);
    if (_containerHeight == value) return;
    _containerHeight = value;
    markNeedsLayout();
  }

  // ---

  double get dx => _dx;
  double _dx;

  set dx(double value) {
    assert(value != null);
    if (_dx == value) return;
    _dx = value;
    markNeedsLayout();
  }

  // ---

  double get dy => _dy;
  double _dy;

  set dy(double value) {
    assert(value != null);
    if (_dy == value) return;
    _dy = value;
    markNeedsLayout();
  }

  // ---

  @override
  void performLayout() {
    //
    if (child != null) {
      child.layout(constraints.loosen(), parentUsesSize: true);
      size = constraints.constrain(const Size(double.infinity, double.infinity));
      alignChild();
    } else {
      size = constraints.constrain(const Size(double.infinity, double.infinity));
    }
  }

  @protected
  void alignChild() {
    assert(child != null);
    assert(!child.debugNeedsLayout);
    assert(child.hasSize);
    assert(hasSize);
    // ---

    final BoxParentData childParentData = child.parentData;
    final Offset containerSize = _toOffset(size);
    final Offset childSize = _toOffset(child.size);

    // 1) Adds touch and alignment.
    if (_touch == Touch.inside)
      childParentData.offset = alignment.alongOffset(containerSize - childSize);
    //
    else if (_touch == Touch.outside)
      childParentData.offset = alignment.alongOffset(containerSize + childSize) - childSize;
    //
    else
      throw AssertionError();

    // 2) Adds dx and dy.
    childParentData.offset += Offset(dx, dy);

    // 3) Adds childWidth and childHeight.
    childParentData.offset += Offset(
      child.size.width * childWidth,
      child.size.height * childHeight,
    );

    // 3) Adds containerWidth and containerHeight.
    childParentData.offset += Offset(
      size.width * containerWidth,
      size.height * containerHeight,
    );
  }

  Offset _toOffset(Size size) => Offset(size.width, size.height);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
